# -*- coding: utf-8 -*-

"""
A package of classes I find myself constantly writing
for separate projects. This is public mostly for my
own convenience and use, although others are free to
use and contribute, of course!

Source:
https://github.com/cmyui/cmyui_pkg

:copyright: (c) 2020 cmyui
:license: MIT, see LICENSE for details.
"""

__title__ = 'cmyui'
__author__ = 'cmyui'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020 cmyui'
__version__ = '1.0.5'

from .connection import *
from .mysqlconn import *
from .packetstream import *
from .server import *
from .types import *
from .version import *



# -*- coding: utf-8 -*-

import socket
from typing import Generator, Tuple, Any

from .types import Address
from .connection import Connection

__all__ = (
    'Server',
)

class Server:
    __slots__ = ('sock_family', 'sock_type', 'listening')

    def __init__(self, socket_family: int = socket.AF_INET,
                 socket_type: int = socket.SOCK_STREAM) -> None:
        if socket_family == socket.AF_UNIX:
            from os import name
            if name == 'nt': raise ValueError(
                'UNIX sockets can only be used on UNIX operating systems.')

        self.sock_family = socket_family
        self.sock_type = socket_type
        self.listening = False

    def __enter__(self) -> None:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def listen(self, addr: Address = ('127.0.0.1', 5001),
               max_conns: int = 5) -> Generator[Connection, None, None]:

        if using_unix := self.sock_family == socket.AF_UNIX:
            if not isinstance(addr, str):
                raise TypeError('Addr must be a string for UNIX sockets.')

            # Remove unix socket if it already exists.
            from os import path, remove
            if path.exists(addr):
                remove(addr)
        else:
            from itertools import izip
            if not isinstance(addr, tuple) or len(addr) != 2 \
            or not all(isinstance(i, t) for i, t in izip(addr, (str, int))):
                raise TypeError(
                    'Addr must be a tuple of (ip: str, port: int) for INET sockets.')

        with socket.socket(self.sock_family, self.sock_type) as s:
            s.bind(addr)

            if using_unix:
                from os import chmod
                chmod(addr, 0o777)

            self.listening = True
            s.listen(max_conns)

            while self.listening:
                yield Connection(*s.accept())

# Simple example usage.
if __name__ == '__main__':
    from socket import AF_UNIX, SOCK_STREAM
    with Server(socket.AF_UNIX, socket.SOCK_STREAM) as s:
        for conn in s.listen('/tmp/gulag.sock', 5):
            print(conn)



# -*- coding: utf-8 -*-

from enum import IntEnum, unique, auto
from typing import Any, Dict, Tuple, Union

__all__ = (
    'Address',
    'DataType',
    'SQLParams',
    'SQLResult'
)

@unique
class DataType(IntEnum):
    pad = 0
    i8 = auto()
    u8 = auto()
    i16 = auto()
    u16 = auto()
    i32 = auto()
    u32 = auto()
    i64 = auto()
    u64 = auto()
    f32 = auto()
    f64 = auto()

    # 'advanced' types
    ''' TODO
    i16_list = auto()
    u16_list = auto()
    i32_list = auto()
    u32_list = auto()
    i64_list = auto()
    u64_list = auto()
    f32_list = auto()
    f64_list = auto()
    string = auto()
    '''

    def __str__(self) -> str:
        if self.value > self.f64:
            raise Exception(
                'Tried to use an advanced type with __str__.')

        return 'xbBhHiIqQfd'[self.value]

# Will be (host: str, port: int) if INET,
# or (sock_dir: str) if UNIX.
Address = Union[Tuple[str, int], str]

SQLParams = Tuple[Union[int, float, str]]
SQLResult = Dict[str, Union[int, float, str]]



# -*- coding: utf-8 -*-

from struct import (
    pack as st_pack,
    unpack as st_unpack
)
from typing import Any, Tuple

from .types import DataType

__all__ = (
    'PacketStream',
    #'BanchoPacketStream'
)

class PacketStream:
    __slots__ = ('data',)

    def __init__(self, data: bytes = None) -> None:
        if data: # Reading
            self.data = data
        else: # Writing
            self.data = bytearray()

    @staticmethod
    def create_fmtstr(types: Tuple[DataType],
                      order: str = '<') -> str:
        return f"{order}{''.join(str(t) for t in types)}"

    def read(self, types: Tuple[DataType]) -> Tuple[Any]:
        return st_unpack(self.create_fmtstr(types, '<'), self.data)

    def write(self, params: Tuple[Tuple[Any, DataType]]) -> None:
        params, types = zip(*params)
        fmt = self.create_fmtstr(types, '<')
        self.data.extend(st_pack(fmt, *params))

#class BanchoPacketStream(PacketStream):
#    pass



# -*- coding: utf-8 -*-

from socket import socket
from collections import defaultdict
from typing import Final, Dict

from .types import Address

__all__ = (
    'http_statuses',
    'Connection',
    'Request',
    'Response'
)

http_statuses: Final[Dict[int, str]] = {
    # Informational
    100: 'CONTINUE',
    101: 'SWITCHING PROTOCOLS',
    102: 'PROCESSING',

    # Success
    200: 'OK',

    # Redirection
    307: 'TEMPORARY REDIRECT',
    308: 'PERMANENT REDIRECT',

    # Client Error
    400: 'BAD REQUEST',
    401: 'UNAUTHORIZED',
    402: 'PAYMENT REQUIRED',
    403: 'FORBIDDEN',
    404: 'NOT FOUND',

    # Server Error
    500: 'INTERNAL SERVER ERROR',
    501: 'NOT IMPLEMENTED',
    502: 'BAD GATEWAY',
    503: 'SERVICE UNAVAILABLE',
    504: 'GATEWAY TIMEOUT'
}

class Request:
    __slots__ = ('headers', 'body', 'data', 'cmd',
                 'uri', 'httpver', 'args', 'files')

    def __init__(self, data: bytes):
        self.data = data

        self.headers = []
        self.body = b''
        self.cmd = ''
        self.uri = ''
        self.httpver = 0.0

        self.args = defaultdict(lambda: None)
        self.files = defaultdict(lambda: None)

        self.parse_http_request()

    def parse_http_request(self) -> None:
        # Retrieve http request line from content.
        req_line, after_req_line = self.data.split(b'\r\n', 1)
        self.cmd, full_uri, _httpver = req_line.decode().split(' ')

        if len(_httpver) != 8 \
        or not _httpver[5:].replace('.', '', 1).isnumeric():
            raise Exception('Invalid HTTP version.')

        self.httpver = float(_httpver[5:])

        # Split into headers & body
        s = after_req_line.split(b'\r\n\r\n', 1)
        self.headers = defaultdict(lambda: None, {
            k: v.lstrip() for k, v in (
                h.split(':') for h in s[0].decode().split('\r\n')
            )
        })
        self.body = s[1]
        del s

        # Parse our request for arguments.
        # Method varies on the http command.
        if self.cmd == 'GET':
            # If our request has arguments, parse them.
            if (p_start := full_uri.find('?')) != -1:
                self.uri = full_uri[:p_start]
                for k, v in (i.split('=') for i in full_uri[p_start + 1:].split('&')):
                    self.args.update({k: v})
            else:
                self.uri = full_uri
        elif self.cmd == 'POST':
            self.uri = full_uri

            if not (ct := self.headers['Content-Type']) \
            or not ct.startswith('multipart/form-data'):
                return # Non-multipart POST

            # Parse multipartform data into args.
            # Very sketch method, i'm not a fan of multipart forms..
            boundary = ct.split('=')[1].encode()

            for form_part in self.body.split(b'--' + boundary)[1:]:
                # TODO len checks? will think abt it
                if form_part[:2] != b'\r\n':
                    continue

                param_lines = form_part[2:].split(b'\r\n', 3)
                if len(param_lines) < 1:
                    raise Exception('Malformed line in multipart form-data.')
                    #continue

                # Find idx manually
                data_line_idx = 0
                for idx, l in enumerate(param_lines):
                    if not l:
                        data_line_idx = idx + 1
                        break

                if data_line_idx == 0:
                    continue

                # XXX: type_line is currently unused, but it
                # can contain the content-type of the param,
                # such as `Content-Type: image/png`.
                attrs_line, type_line = (s.decode() for s in param_lines[:2])

                if not attrs_line.startswith('Content-Disposition: form-data'):
                    continue

                if ';' not in attrs_line:
                    continue

                # Line has attributes in `k="v"` fmt, each
                # delimited by ` ;`. We split by `;` and
                # lstrip the key to allow for a `;` delimiter.
                attrs = {k.lstrip(): v[1:-1] for k, v in (
                    a.split('=', 1) for a in attrs_line.split(';')[1:]
                )}

                if 'filename' in attrs:
                    self.files.update({attrs['filename']: param_lines[data_line_idx]})
                elif 'name' in attrs:
                    self.args.update({attrs['name']: param_lines[data_line_idx]})
                else:
                    continue

                # Link all other attrs to their respective values.
                for k, v in attrs.items():
                    if k != 'name':
                        self.args.update({k: v})
        else:
            raise Exception(f'Unsupported HTTP command: {self.cmd}')

class Response:
    __slots__ = ('sock', 'headers')

    def __init__(self, sock: socket) -> None:
        self.sock = sock
        self.headers = []

    def add_header(self, header: str) -> None:
        self.headers.append(header)

    def send(self, data: bytes, code: int = 200) -> None:
        # Insert HTTP response line & content
        # length at the beginning of the headers.
        self.headers.insert(0, f'HTTP/1.1 {code} {http_statuses[code]}')
        self.headers.insert(1, f'Content-Length: {len(data)}')
        try:
            self.sock.send('\r\n'.join(self.headers).encode() + b'\r\n\r\n' + data)
        except BrokenPipeError:
            print('\x1b[1;91mWARN: Connection pipe broken.\x1b[0m')

class Connection: # will probably end up removing addr?
    __slots__ = ('request', 'response', 'addr')

    def __init__(self, sock: socket, addr: Address) -> None:
        self.request = Request(self.read_data(sock))
        self.response = Response(sock)
        self.addr = addr

    @staticmethod
    def read_data(sock: socket, ch_size: int = 1024) -> bytes:
        data = sock.recv(ch_size)

        while (l := len(data)) % ch_size == 0:
            data += sock.recv(ch_size)
            if l == len(data):
                # No growth
                break

        return data
